/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"

#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>



int 
setup_socket(void)
{
	struct sockaddr_in MA;
	struct ip_mreq mreq;
	int on = 1, fl;
	int sock;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (-1 == sock)
		error("cannot create socket: %s\n",strerror(errno));
	fl = fcntl(sock,F_GETFL,0);
	fcntl(sock,F_SETFL,fl | O_NONBLOCK);
	if (-1 == setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on)))
		error("cannot set socket to reuse address: %s\n",strerror(errno));
	bzero(&MA,sizeof(MA));
	MA.sin_family = AF_INET;
	MA.sin_port = htons(7536);
	if (-1 == inet_pton(AF_INET, "224.0.0.75", &MA.sin_addr))
		error("cannot set multicast address: %s\n",strerror(errno));
	if (-1 == bind(sock,(struct sockaddr *)&MA,sizeof(MA)))
		error("cannot bind socket: %s\n",strerror(errno));

	memcpy(&mreq.imr_multiaddr.s_addr,(void *)&MA.sin_addr,sizeof(struct in_addr));
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	if (-1 == setsockopt(sock,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)))
		error("could not join multicast group: %s\n",strerror(errno));
	return sock;
}


static void
parse_dev(char *dev, char *m, sysstat_t *s)
{
	char key[4173], value[4173];
	fsstat_t *l;
	fsstat_t *fs = calloc(1,sizeof(fsstat_t));
	if (s->fs) {
		l = s->fs;
		while (l->next) l = l->next;
		l->next = fs;
	} else {
		s->fs = fs;
	}
	fs->dev = strdup(dev);
	//dbug("parse_dev(%s,%s,%p)\n",dev,m,s);
	assert(m && (*m == '\t'));
	++m;
	while (m && (2 == sscanf(m,"%[^\t]\t%[^\t\n]",key,value))) {
		//dbug("parse_dev(): key = \"%s\", value = \"%s\"\n",key,value);
		if (0 == strcmp("mnt",key)) {
			fs->mnt = strdup(value);
		} else if (0 == strcmp("fs",key)) {
			if (strlen(value) > 7) 
				warn("protocol error: %s\t%s\n",key,value);
			else
				strcpy(fs->typ,value);
		} else if (0 == strcmp("fr",key)) {
			if (sscanf(value,"%llu",&fs->free) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("bl",key)) {
			if (sscanf(value,"%llu",&fs->blocks) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else
			warn("unkown fs subfield: %s\t%s\n",key,value);
		m = strchr(m,'\t');
		assert(m);
		++m;
		m = strchr(m,'\t');
		if (m == 0)
			return;
		++m;
	}
	if (fs->mnt == 0)
		fs->mnt = strdup("");
}


static void
parse_user(char *uid, char *m, sysstat_t *s)
{
	char key[4173], value[4173];
	userstat_t *l = 0;
	userstat_t *us = calloc(1,sizeof(userstat_t));
	if (1 != sscanf(uid,"%lu",&us->uid)) {
		warn("protocol error: uid\t%s\n",uid);
		free(us);
		return;
	}
	if (s->us) {
		l = s->us;
		while (l->next) l = l->next;
		l->next = us;
	} else {
		s->us = us;
	}
	assert(m && (*m == '\t'));
	++m;
	while (m && (2 == sscanf(m,"%[^\t]\t%[^\t\n]",key,value))) {
		//dbug("parse_user(): key = \"%s\", value = \"%s\"\n",key,value);
		if (0 == strcmp("user",key)) {
			us->username = strdup(value);
		} else if (0 == strcmp("N",key)) {
			if (sscanf(value,"%lu",&us->nproc) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("L",key)) {
			if (sscanf(value,"%lu",&us->nlwp) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("O",key)) {
			if (sscanf(value,"%lu",&us->oncpu) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("R",key)) {
			if (sscanf(value,"%lu",&us->runnable) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("S",key)) {
			if (sscanf(value,"%lu",&us->sleep) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("T",key)) {
			if (sscanf(value,"%lu",&us->stopped) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("Z",key)) {
			if (sscanf(value,"%lu",&us->zombie) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("sec",key)) {
			if (sscanf(value,"%ld",&us->time) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("RSS",key)) {
			if (sscanf(value,"%lu",&us->rss) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("SIZE",key)) {
			if (sscanf(value,"%lu",&us->size) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("%C",key)) {
			if (sscanf(value,"%lf",&us->pctcpu) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else if (0 == strcmp("%M",key)) {
			if (sscanf(value,"%lf",&us->pctmem) != 1) 
				warn("protocol error: %s\t%s\n",key,value);
		} else
			warn("unkown user subfield: %s\t%s\n",key,value);
		m = strchr(m,'\t');
		assert(m);
		++m;
		m = strchr(m,'\t');
		if (m == 0)
			return;
		++m;
	}
	if (us->username == 0)
		us->username = strdup("<unknown>");
}


sysstat_t *
receive_stats(int sock)
{
	char msg[4172+1], *m;	/* maximum size of one datagram packet plus \0 */
	/* make key and value big enough, so that buffer overflow attacks have no chance */
	char key[4172+1],value[4172+1];
	int num;
	struct sockaddr_in from;
	socklen_t len;
	sysstat_t *s;

	num = recvfrom(sock,&msg,sizeof(msg),0,(struct sockaddr *)&from,&len);
	if ((num == -1) && (errno == EWOULDBLOCK))
		return 0;
	if (num == -1) {
		error("cannot read from socket: %s\n",strerror(errno));
	}
	msg[4172] = '\0';
	/*dbug("received packet:\n%s\n",msg);*/
	m = msg;
	if (sscanf(m,"%[^\t]\t%[^\n]\n",key,value) != 2) {
		warn("unknown protocol\n");
		return 0;
	}
	if (strcmp(key,"sysstat") || strcmp(value,"1")) {
		warn("invalid protcol\n");
		return 0;
	}
	m = strchr(m,'\n');
	assert(m);
	++m;
	s = calloc(1,sizeof(sysstat_t));
	assert(s);
	while (m && (sscanf(m,"%[^\t]\t%[^\t\n]",key,value) == 2)) {
		//dbug("key: \"%s\", value: \"%s\"\n",key,value);
		if (strcmp("NCPU",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->Ncpus))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("OCPU",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->Ocpus))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Host",key) == 0) {
			if (strlen(value) > sizeof(s->Hostinfo))
				warn("protocol error: %s\t%s\n",key,value);
			else
				strcpy(s->Hostinfo,value);
		} else if (strcmp("Itv",key) == 0) {
			/* currently ignored */
		} else if (strcmp("TS",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->TS))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Disk",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->Devices))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("User",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->UserP))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Kernel",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->KernelP))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Idle",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->IdleP))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Load1",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->loadavg[0]))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Load5",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->loadavg[1]))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Load15",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->loadavg[2]))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("intr",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->IntrR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("intrblk",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->IntrBlkR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("pswitch",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->PswitchR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("inv_swtch",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->InvSwtchR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("cpumigrate",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->MigrateR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("xcall",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->XcallsR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("call",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->SyscallR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("exec",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->SysexecR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("fork",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->SysforkR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("read",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->SysreadR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("write",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->SyswriteR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Ram",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->Physmem))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Free",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->Pagesfree))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Locked",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->Pageslocked))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("SwapTotal",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->SwapTotal))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("SwapFree",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->SwapFree))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("scan",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->ScanR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("pgswapin",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->PgswapinR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("pgswapout",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->PgswapoutR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("Storage",key) == 0) {
			if (1 != sscanf(value,"%llu",&s->Storage))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("FsFree",key) == 0) {
			if (1 != sscanf(value,"%llu",&s->FsFree))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("HErr",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->HardErrors))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("SErr",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->SoftErrors))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("IpIErr",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->IpInErrs))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("IpIDis",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->IpInDiscards))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("IpODis",key) == 0) {
			if (1 != sscanf(value,"%lu",&s->IpOutDiscards))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("IpI",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->IpInR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("IpO",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->IpOutR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("NfsI",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->NfsIR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("NfsO",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->NfsOR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("DiskI",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->DiskIR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("DiskO",key) == 0) {
			if (1 != sscanf(value,"%lf",&s->DiskOR))
				warn("protocol error: %s\t%s\n",key,value);
		} else if (strcmp("uid",key) == 0) {
			char *eof = strchr(m,'\n');
			assert(eof);
			*eof = '\0';
			m = strchr(m,'\t');
			assert(m);
			++m;
			parse_user(value,strchr(m,'\t'),s);
			*eof = '\n';
		} else if (strcmp("dev",key) == 0) {
			char *eof = strchr(m,'\n');
			assert(eof);
			*eof = '\0';
			m = strchr(m,'\t');
			assert(m);
			++m;
			parse_dev(value,strchr(m,'\t'),s);
			++s->numfs;
			*eof = '\n';
		} else {
			warn("unknown key: %s\n",key);
		}
		m = strchr(m,'\n');
		assert(m);
		++m;
	}
	return s;
}

// vim:tw=0
