/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"

#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>



static char
	Msg[4172+1],	/* maximum size of one datagram packet plus
			   terminating \0 */
	*M = Msg;
static size_t
	S;
static const int
	Port = 7536;



static char *
per2str(double d)
{
	static char str[7];
	assert(d <= 1E2);
	if (d < 0) {
		sprintf(str,"%.0f",d);
	} else if (d < 2E-2) {
		sprintf(str,"%f",d);
	} else if (d < 9.995) {
		sprintf(str,"%.2f",d);
	} else if (d < 99.95) {
		sprintf(str,"%.1f",d);
	} else
		sprintf(str,"%f",rint(d));
	return str;
}


static char *
dbl2str(double d)
{
	static char str[16];
	double r = rint(d);
	if (d < 0) {
		sprintf(str,"%.0f",r);
	} else if (d < 2E-2) {
		strcpy(str,"0");
	} else if (d < 1E1) {
		sprintf(str,"%.2f",d);
	} else if (r < 1E2) {
		sprintf(str,"%.1f",d);
	} else {
		sprintf(str,"%.0f",r);
	}
	return str;
}


static char *
kb2str(long long ll)
{
	static char str[16];
	sprintf(str,"%.0f",(double)ll);
	return str;
}


static void
add_int(char *f, long long v, char c)
{
	size_t c0,c1;

	if (S == 0)
		return;
	assert(f);
	c0 = strlcpy(M,f,S);
	if (c0 >= S)
		return;
	S -= c0;
	M += c0;
	*M++ = '\t';
	--S;
	c1 = snprintf(M,S,"%lld",v);
	if (c1 > S)
		return;
	S -= c1;
	M += c1;
	*M++ = c;
	--S;
}


static void
add_str(char *f, char *v, char c)
{
	size_t c0,c1;

	if (S == 0)
		return;
	assert(f && v);
	c0 = strlcpy(M,f,S);
	if (c0 >= S)
		return;
	S -= c0;
	M += c0;
	*M++ = '\t';
	--S;
	c1 = strlcpy(M,v,S);
	if (c1 > S)
		return;
	S -= c1;
	M += c1;
	*M++ = c;
	--S;
}


void send_to_network(sysstat_t *s)
{
	static int sock = -1;
	static struct sockaddr_in MA;
	fsstat_t *fs;
	userstat_t *us;

	if (-1 == sock) {
		sock = socket(AF_INET, SOCK_DGRAM, 0);
		if (-1 == sock)
			error("could not create socket: %s\n",strerror(errno));
		bzero(&MA,sizeof(MA));
		MA.sin_family = AF_INET;
		MA.sin_port = htons(Port);
		if (-1 == inet_pton(AF_INET, "224.0.0.75", &MA.sin_addr))
			error("could not set multicast address: %s\n",strerror(errno));
	}
	M = Msg;
	S = sizeof(Msg)/sizeof(char) - 1;

	add_str("sysstat","1",'\n');
	add_int("Itv",Interval,'\n');
	add_str("Host",s->Hostinfo,'\n');
	add_int("TS",s->TS,'\n');
	add_int("NCPU",s->Ncpus,'\n');
	add_int("OCPU",s->Ocpus,'\n');
	add_int("Disk",s->Devices,'\n');
	add_str("User",per2str(s->UserP),'\n');
	add_str("Kernel",per2str(s->KernelP),'\n');
	add_str("Idle",per2str(s->IdleP),'\n');
	add_str("Load1",dbl2str(s->loadavg[0]),'\n');
	add_str("Load5",dbl2str(s->loadavg[1]),'\n');
	add_str("Load15",dbl2str(s->loadavg[2]),'\n');
	add_str("intr",dbl2str(s->IntrR),'\n');
	add_str("intrblk",dbl2str(s->IntrBlkR),'\n');
	add_str("pswitch",dbl2str(s->PswitchR),'\n');
	add_str("inv_swtch",dbl2str(s->InvSwtchR),'\n');
	add_str("cpumigrate",dbl2str(s->MigrateR),'\n');
	add_str("xcall",dbl2str(s->XcallsR),'\n');
	add_str("Ram",kb2str(s->Physmem),'\n');
	add_str("Free",kb2str(s->Pagesfree),'\n');
	add_str("Locked",kb2str(s->Pageslocked),'\n');
	add_str("SwapTotal",kb2str(s->SwapTotal),'\n');
	add_str("SwapFree",kb2str(s->SwapFree),'\n');
	add_str("scan",dbl2str(s->ScanR),'\n');
	add_str("pgswapin",dbl2str(s->PgswapinR),'\n');
	add_str("pgswapout",dbl2str(s->PgswapoutR),'\n');
	add_str("call",dbl2str(s->SyscallR),'\n');
	add_str("exec",dbl2str(s->SysexecR),'\n');
	add_str("fork",dbl2str(s->SysforkR),'\n');
	add_str("read",dbl2str(s->SysreadR),'\n');
	add_str("write",dbl2str(s->SyswriteR),'\n');
	add_str("IpI",dbl2str(s->IpInR),'\n');
	add_str("IpO",dbl2str(s->IpOutR),'\n');
	add_str("Storage",kb2str(s->Storage),'\n');
	add_str("FsFree",kb2str(s->FsFree),'\n');
	add_str("DiskI",kb2str(s->DiskIR/1024),'\n');
	add_str("DiskO",kb2str(s->DiskOR/1024),'\n');
	add_str("NfsI",kb2str(s->NfsIR/1024),'\n');
	add_str("NfsO",kb2str(s->NfsOR/1024),'\n');
	add_int("IpIErr",s->IpInErrs,'\n');
	add_int("IpIDis",s->IpInDiscards,'\n');
	add_int("IpODis",s->IpOutDiscards,'\n');
	add_int("HErr",s->HardErrors,'\n');
	add_int("SErr",s->SoftErrors,'\n');
	
	fs = s->fs;
	us = s->us;
	while ((fs || us) && S) {
		if (us) {
			unsigned long long s;
			add_int("uid",us->uid,'\t');
			add_str("user",us->username,'\t');
			add_int("L",us->nlwp,'\t');
			if (us->oncpu)
				add_int("O",us->oncpu,'\t');
			if (us->sleep)
				add_int("S",us->sleep,'\t');
			if (us->runnable)
				add_int("R",us->runnable,'\t');
			if (us->zombie)
				add_int("Z",us->zombie,'\t');
			if (us->stopped)
				add_int("T",us->stopped,'\t');
			add_str("SIZE",kb2str(us->size),'\t');
			add_str("RSS",kb2str(us->rss),'\t');
			if (us->pctmem >= 2E-2)
				add_str("%M",per2str(us->pctmem),'\t');
			if (us->pctcpu >= 2E-2)
				add_str("%C",per2str(us->pctcpu),'\t');
			s = us->time.tv_sec;
			s += (int)rint((double)us->time.tv_nsec/1000000000);
			assert(s >= 0);
			if (s)
				add_int("sec",s,'\t');
			add_int("N",us->nproc,'\n');
			us = us->next;
		}
		if (fs) {
			add_str("dev",fs->dev,'\t');
			add_str("fs",fs->typ,'\t');
			add_str("mnt",fs->mnt,'\t');
			add_str("fr",kb2str(fs->free),'\t');
			add_str("bl",kb2str(fs->blocks),'\n');
			fs = fs->next;
		}
	}
	if (S == 0)
		warn("cannot xmit all data\n");
	*M = 0;
	if (-1 == sendto(sock, Msg, M - Msg + 1, 0, (const struct sockaddr *) &MA, sizeof(MA))) {
		warn("could not multicast message: %s",strerror(errno));
		abort();
	}
}
