/*
 * Copyright 2006, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "log.h"
#include "network.h"
#include "sysstat.h"

int Fslines = 5;
unsigned Interval = 5;

static sysstat_t
	*SS = 0, *Current = 0;


static void
dealloc(sysstat_t *s)
{
	userstat_t *us;
	fsstat_t *fs;
	if (s == 0)
		return;
	us = s->us;
	while (us) {
		userstat_t *n = us->next;
		free(us->username);
		free(us);
		us = n;
	}
	fs = s->fs;
	while (fs) {
		fsstat_t *n = fs->next;
		free(fs->dev);
		free(fs->mnt);
		free(fs);
		fs = n;
	}
	free(s);
}


static void
add_stats(sysstat_t *s)
{
	sysstat_t *i = SS;
	assert((s != i) && (s->next == 0) && (s->prev == 0));
	/* search, if we already have an entry for this host */
	while (i && s) {
		assert(i->next && i->prev);
		if (0 == strcmp(i->Hostinfo,s->Hostinfo)) {
			/* replace this entry */
			if (i->prev != i) {
				assert(i->next != i);
				s->prev = i->prev;
				assert(s->prev->next == i);
				s->prev->next = s;
				s->next = i->next;
				assert(s->next->prev == i);
				s->next->prev = s;
			} else {
				assert(i->next == i);
				s->prev = s;
				s->next = s;
			}
			if (Current == i)
				Current = s;
			if (SS == i)
				SS = s;
			dealloc(i);
			assert(s->prev && s->next);
			return;
		}
		i = i->next;
		if (i == SS)
			break;
	}
	/* new sysstat not already in our list */
	assert(s);
	if (SS) {
		s->next = SS;
		s->prev = SS->prev;
		assert(s->prev->next == SS);
		s->prev->next = s;
		SS->prev = s;
	} else {
		s->next = s;
		s->prev = s;
		SS = s;
		Current = s;
	}
	assert(s->prev && s->next);
}


static void
process_input(int s, int itv)
{
	struct pollfd fds[2];
	char c;
	sysstat_t *st;
	
	fds[0].fd = STDIN_FILENO;
	fds[0].events = POLLIN;
	fds[1].fd = s;
	fds[1].events = POLLIN;
	if (-1 == poll(fds,s == -1 ? 1 : sizeof(fds)/sizeof(struct pollfd),itv)) {
		if (errno != EINTR)
			error("error polling: %s\n",strerror(errno));
		return;
	}
	if (1 == read(STDIN_FILENO,&c,sizeof(char))) {
		switch (c) {
			case 'q':
				exit(EXIT_SUCCESS);
				break;
			case 'h':
				if (Current == 0)
					Current = SS;
				else
					Current = Current->next;
				break;
			case 'H':
				if (Current == 0)
					Current = SS;
				else
					Current = Current->prev;
				break;
			case 'n':
				set_sort_by_nproc();
				break;
			case 'r':
				set_sort_by_rss();
				break;
			case 's':
				set_sort_by_size();
				break;
			case 't':
				set_sort_by_time();
				break;
			case 'u':
				set_sort_by_user();
				break;
			case 'c':
				set_sort_by_pcpu();
				break;
			default:
				;
		}
	}
	if ((s != -1) && (st = receive_stats(s))) {
		assert(st != SS);
		add_stats(st);
	}
}


static void
usage(int r)
{
	printf(
	"sysstat [option] [interval [count]]\n"
	"Options:\n"
	"-h      : display this help screen and exit\n"
	"-c      : write stats to stdout instead of using curses\n"
	"-d      : start sysstat as daemon, multicasting stats\n"
	"-f<num> : show at least num lines of filesystem information\n"
	"-l      : start sysstat in local-mode (i.e. without network support)\n"
	"-r	 : sort user/process summary by rss\n"
	"-s      : sort user/process summary by size\n"
	"-t      : sort user/process summary by %%CPU time\n"
	"-u      : sort user/process summary by username\n"
	"-v      : operate in verbose mode\n"
	"-V      : display version number and exit\n"
	);
	exit(r);
}


static void
version(void)
{
	printf(
	"sysstat version 20060511\n"
	"Copyright 2006, Thomas Maier-Komor\n"
	"This software is licensed under the GPLv2. See file COPYING for details.\n"
	);
	exit(EXIT_SUCCESS);
}


extern int init_curses(void);
extern void display_curses(sysstat_t *);
extern void display_stdout(sysstat_t *);

int
main(int argc, char **argv)
{
	unsigned curses = 1, daemon = 0, multicast = 0, localmode = 0;
	int c, count = -1, pid, sock;
	extern char *optarg;
	FILE *pidf;

	while ((c = getopt(argc, argv, "cdf:hlrstuvV")) != -1) {
		switch(c) {
		default:
			assert(0);
			break;
		case 'c':
			curses = 0;
			break;
		case 'd':
			daemon = 1;
			multicast = 1;
			break;
		case 'f':
			if (1 != sscanf(optarg,"%u",&Fslines))
				usage(EXIT_FAILURE);
			break;
		case 'h':
			usage(EXIT_SUCCESS);
			break;
		case 'l':
			localmode = 1;
			break;
		case 'r':
			set_sort_by_rss();
			break;
		case 's':
			set_sort_by_size();
			break;
		case 't':
			set_sort_by_time();
			break;
		case 'u':
			set_sort_by_user();
			break;
		case 'v':
			Verbose = 1;
			break;
		case 'V':
			version();
			break;
		case '?':
			usage(EXIT_FAILURE);
			break;
		}
	}
	if (optind < argc) {
		if (1 != sscanf(argv[optind],"%u",&Interval))
			usage(EXIT_FAILURE);
		dbug("interval set to %u\n",Interval);
		++optind;
	}
	if (optind < argc) {
		if (1 != sscanf(argv[optind],"%d",&count) || (count <= 0))
			usage(EXIT_FAILURE);
		dbug("count set to %u\n",count);
		++optind;
	}
	if (optind < argc) {
		usage(EXIT_FAILURE);
	}
	if (strcmp(argv[0], "sysstatd") == 0) {
		multicast = 1;
		daemon = 1;
	}
	if (localmode) {
		sysstat_t SS;
		SS.next = &SS;
		SS.prev = &SS;
		init_localstat(&SS);
		if (curses)
			init_curses();
		for (;count;--count) {
			get_local_data(&SS);
			if (curses) {
				display_curses(&SS);
			} else {
				display_stdout(&SS);
			}
			process_input(-1,Interval);
		}
	}
	pidf = fopen("/tmp/sysstat.pid","r");
	if (pidf) {
		char pn[] = "/proc/\0\0\0\0\0\0",*p;
		struct stat st;
		info("/tmp/sysstat.pid exists\n");
		if (1 != fscanf(pidf,"%d",&pid)) {
			warn("corrupt pid file /tmp/systat.pid - unlinking it\n");
			if (-1 == unlink("/tmp/sysstat.pid"))
				warn("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
			multicast = 1;
		} else {
			p = pn + 6;
			sprintf(p,"%d",pid);
			info("stating %s\n",pn);
			if (-1 == stat(pn,&st)) {
				info("unlinking stale pid file /tmp/sysstat.pid\n");
				if (-1 == unlink("/tmp/sysstat.pid"))
					warn("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
				multicast = 1;
			} else {
				info("daemon already running with pid %d\n",pid);
				if (daemon)
					exit(EXIT_FAILURE);
			}
		}
	} else
		multicast = 1;
        if (multicast) {
		pid_t pid;
		if (0 == fork()) {
			if (pid = fork()) {
				FILE *pidf;
				pidf = fopen("/tmp/sysstat.pid","w");
				if (pidf == 0)
					error("cannot create pid file: %s\n",strerror(errno));
				fprintf(pidf,"%d\n",pid);
				fclose(pidf);
				chmod("/tmp/sysstat.pid",0666);
				exit(EXIT_SUCCESS);
			} else
				rundaemon(daemon,count,Interval);
		}
	}
	if (daemon)
		exit(EXIT_SUCCESS);

	sock = setup_socket();
	if (-1 == fcntl(STDIN_FILENO,F_SETFL, O_NONBLOCK))
		error("Cannot set input file to non-blocking: %s\n",strerror(errno));
	if (curses)
		curses = init_curses();
	for (;;) {
		if (curses) {
			display_curses(Current);
		} else {
			display_stdout(Current);
		}
		process_input(sock,-1);
	}
	/*return 0; 	-- lint wants it, cc doesn't... */
}

// vim:tw=0
